/*
Author: Cao Thanh Tung
Date: 21/01/2010

File Name: main.cpp

===============================================================================

Copyright (c) 2010, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/pba.html

If you use PBA and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "pba\pba3D.h"

//**********************************************************************************************
//* Input parameters
//**********************************************************************************************
int fboSize     = 64;
int nVertices   = 100; 

int phase1Band  = 1; 
int phase2Band	= 1; 
int phase3Band	= 2; 

#define TOID(x, y, z, w)    ((z) * (w) * (w) + (y) * (w) + (x))

typedef struct {
    double totalDistError, maxDistError; 
    int errorCount; 
} ErrorStatistics; 

//**********************************************************************************************
//* Global Vars
//**********************************************************************************************
int *inputPoints, *inputVoronoi, *outputVoronoi; 
ErrorStatistics pba; 

//**********************************************************************************************
//* Random Point Generator
//**********************************************************************************************
// Random number generator, obtained from http://oldmill.uchicago.edu/~wilder/Code/random/
unsigned long z, w, jsr, jcong; // Seeds
void randinit(unsigned long x_) 
{ z =x_; w = x_; jsr = x_; jcong = x_; }
unsigned long znew() 
{ return (z = 36969 * (z & 0xfffful) + (z >> 16)); }
unsigned long wnew() 
{ return (w = 18000 * (w & 0xfffful) + (w >> 16)); }
unsigned long MWC()  
{ return ((znew() << 16) + wnew()); }
unsigned long SHR3()
{ jsr ^= (jsr << 17); jsr ^= (jsr >> 13); return (jsr ^= (jsr << 5)); }
unsigned long CONG() 
{ return (jcong = 69069 * jcong + 1234567); }
unsigned long rand_int()         // [0,2^32-1]
{ return ((MWC() ^ CONG()) + SHR3()); }
double random()     // [0,1)
{ return ((double) rand_int() / (double(ULONG_MAX)+1)); }

// Generate input points
void generateRandomPoints(int texSize, int nPoints)
{	
    int tx, ty, tz, id; 

    randinit(0);

    for (int i = 0; i < texSize * texSize * texSize; i++)
        inputVoronoi[i] = MARKER; 

	for (int i = 0; i < nPoints; i++)
	{
        do { 
            tx = int(random() * texSize); 
            ty = int(random() * texSize); 
            tz = int(random() * texSize); 
            id = TOID(tx, ty, tz, texSize); 
		} while (inputVoronoi[id] != MARKER); 

        inputVoronoi[id] = ENCODE(tx, ty, tz); 
        inputPoints[i] = ENCODE(tx, ty, tz);  
    }	
}

/**********************************************************************************************
 * Deinitialization
 **********************************************************************************************/
void deinitialization()
{
    pba3DDeinitialization(); 

    free(inputPoints); 
    free(inputVoronoi); 
    free(outputVoronoi); 
}

/**********************************************************************************************
 * Initialization
 **********************************************************************************************/
void initialization()
{
    pba3DInitialization(fboSize); 

    inputPoints     = (int *) malloc(nVertices * sizeof(int)); 
    inputVoronoi    = (int *) malloc(fboSize * fboSize * fboSize * sizeof(int)); 
    outputVoronoi   = (int *) malloc(fboSize * fboSize * fboSize * sizeof(int)); 
}

/**********************************************************************************************
* Verify the output Voronoi Diagram
 **********************************************************************************************/
void compareResult(ErrorStatistics *e) 
{
    e->totalDistError = 0.0; 
    e->maxDistError = 0.0; 
    e->errorCount = 0; 

    int dx, dy, dz, nx, ny, nz; 
    double dist, myDist, correctDist, error;

    for (int i = 0; i < fboSize; i++)
        for (int j = 0; j < fboSize; j++) 
            for (int k = 0; k < fboSize; k++) {
                int id = TOID(i, j, k, fboSize); 
                DECODE(outputVoronoi[id], nx, ny, nz); 

                dx = nx - i; dy = ny - j; dz = nz - k; 
                correctDist = myDist = dx * dx + dy * dy + dz * dz; 

                for (int t = 0; t < nVertices; t++) {
                    DECODE(inputPoints[t], nx, ny, nz); 
                    dx = nx - i; dy = ny - j; dz = nz - k; 
                    dist = dx * dx + dy * dy + dz * dz; 

                    if (dist < correctDist)
                        correctDist = dist; 
                }

                if (correctDist != myDist) {
                    error = fabs(sqrt(myDist) - sqrt(correctDist)); 

                    e->errorCount++; 
                    e->totalDistError += error; 

                    if (error > e->maxDistError)
                        e->maxDistError = error; 
                }
            }
}

void printStatistics(ErrorStatistics *e)
{
    double avgDistError = e->totalDistError / e->errorCount; 

    if (e->errorCount == 0)
        avgDistError = 0.0; 

    printf("* Error count           : %i -> %.3f%\n", e->errorCount, 
        (double(e->errorCount) / nVertices) * 100.0);
    printf("* Max distance error    : %.5f\n", e->maxDistError);
    printf("* Average distance error: %.5f\n", avgDistError);
}

/**********************************************************************************************
 * Run the tests
 **********************************************************************************************/
void runTests()
{
    printf("Generate input...\n"); 
    generateRandomPoints(fboSize, nVertices); 

    printf("Running PBA to compute 3D Voronoi Diagram...\n"); 
	pba3DVoronoiDiagram(inputVoronoi, outputVoronoi, phase1Band, phase2Band, phase3Band); 

    printf("Verifying the result...\n"); 
    compareResult(&pba);

	printf("-----------------\n");
	printf("Texture: %dx%dx%d\n", fboSize, fboSize, fboSize);
	printf("Points: %d\n", nVertices);
	printf("-----------------\n");

    printStatistics(&pba); 
}

int main(int argc, char **argv)
{
    initialization(); 
    runTests(); 
    deinitialization(); 

    getchar(); 
		
	return 0;
}