/*
Author: Cao Thanh Tung
Date: 21/01/2010

File Name: pba3DHost.cu

===============================================================================

Copyright (c) 2010, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/pba.html

If you use PBA and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#include <device_functions.h>

#include "pba3D.h"

// Parameters for CUDA kernel executions
#define BLOCKX      32
#define BLOCKY      4
#define BLOCKXY     16

#define INFINITY    0x3ff

/****** Global Variables *******/
int **pbaTextures;   

int pbaMemSize; 
int pbaCurrentBuffer; 
int pbaTexSize;  
int log2Width; 

texture<int> pbaTexColor; 
texture<int> pbaTexLinks; 
texture<short> pbaTexPointer; 

/********* Kernels ********/
#include "pba3DKernel.h"

///////////////////////////////////////////////////////////////////////////
//
// Initialize necessary memory for 3D Voronoi Diagram computation
// - textureSize: The size of the Discrete Voronoi Diagram (width = height)
//
///////////////////////////////////////////////////////////////////////////
void pba3DInitialization(int fboSize)
{
    int tmp = pbaTexSize = fboSize; 
    log2Width = 0; 
    while (tmp > 1) { tmp /= 2; log2Width++; }

	pbaTextures = (int **) malloc(2 * sizeof(int *)); 

	pbaMemSize = pbaTexSize * pbaTexSize * pbaTexSize * sizeof(int); 

	// Allocate 2 textures
	cudaMalloc((void **) &pbaTextures[0], pbaMemSize); 
	cudaMalloc((void **) &pbaTextures[1], pbaMemSize); 
}

///////////////////////////////////////////////////////////////////////////
//
// Deallocate all allocated memory
//
///////////////////////////////////////////////////////////////////////////
void pba3DDeinitialization()
{
	cudaFree(pbaTextures[0]); 
	cudaFree(pbaTextures[1]); 

	free(pbaTextures); 
}

// Copy input to GPU 
void pba3DInitializeInput(int *input)
{
    cudaMemcpy(pbaTextures[0], input, pbaMemSize, cudaMemcpyHostToDevice); 

    // Set Current Source Buffer
	pbaCurrentBuffer = 0;
}

// In-place transpose a cubic texture. 
// Transposition are performed on each XY plane. 
// Point coordinates are also swapped. 
void pba3DTransposeXY(int *texture) {
    dim3 block(BLOCKXY, BLOCKXY); 
    dim3 grid((pbaTexSize / BLOCKXY) * pbaTexSize, pbaTexSize / BLOCKXY); 

    kernelTransposeXY<<< grid, block >>>(texture, log2Width, pbaTexSize - 1); 
}

// Phase 1 of PBA. m1 must divides texture size
// Sweeping are done along the Z axiz. 
void pba3DColorZAxis(int m1) 
{
   	dim3 block = dim3(BLOCKX, BLOCKY); 
    dim3 grid = dim3((pbaTexSize / block.x) * m1, pbaTexSize / block.y); 

    cudaBindTexture(0, pbaTexColor, pbaTextures[pbaCurrentBuffer]); 
    kernelFloodZ<<< grid, block >>>(pbaTextures[1 - pbaCurrentBuffer], pbaTexSize, pbaTexSize / block.x, pbaTexSize / m1); 
    pbaCurrentBuffer = 1 - pbaCurrentBuffer; 

	if (m1 > 1) {
		// Passing information between bands
		cudaBindTexture(0, pbaTexColor, pbaTextures[pbaCurrentBuffer]); 
		kernelPropagateInterband<<< grid, block >>>(pbaTextures[1 - pbaCurrentBuffer], pbaTexSize, pbaTexSize / block.x, pbaTexSize / m1); 

		cudaBindTexture(0, pbaTexLinks, pbaTextures[1 - pbaCurrentBuffer]); 
		kernelUpdateVertical<<< grid, block >>>(pbaTextures[pbaCurrentBuffer], pbaTexSize, pbaTexSize / block.x, pbaTexSize / m1); 
	}
}

// Phase 2 of PBA. m2 must divides texture size. 
// This method work along the Y axis
void pba3DComputeProximatePointsYAxis(int m2) 
{
	int iStack = 1 - pbaCurrentBuffer; 
	int iForward = pbaCurrentBuffer; 

	dim3 block = dim3(BLOCKX, BLOCKY); 
    dim3 grid = dim3((pbaTexSize / block.x) * m2, pbaTexSize / block.y); 

	// Compute proximate points locally in each band
	cudaBindTexture(0, pbaTexColor, pbaTextures[pbaCurrentBuffer]); 
    kernelMaurerAxis<<< grid, block >>>(pbaTextures[iStack], pbaTexSize, pbaTexSize / block.x, pbaTexSize / m2); 

	// Construct forward pointers
	cudaBindTexture(0, pbaTexLinks, pbaTextures[iStack]); 
	kernelCreateForwardPointers<<< grid, block >>>((short *) pbaTextures[iForward], pbaTexSize, pbaTexSize / block.x, pbaTexSize / m2); 

	cudaBindTexture(0, pbaTexPointer, pbaTextures[iForward]); 

	// Repeatly merging two bands into one
	for (int noBand = m2; noBand > 1; noBand /= 2) {
		grid = dim3((pbaTexSize / block.x) * (noBand / 2), pbaTexSize / block.y); 
		kernelMergeBands<<< grid, block >>>(pbaTextures[iStack], 
			(short *) pbaTextures[iForward], pbaTexSize, pbaTexSize / block.x, pbaTexSize / noBand); 
	}

	cudaUnbindTexture(pbaTexLinks); 
	cudaUnbindTexture(pbaTexColor); 
	cudaUnbindTexture(pbaTexPointer); 
}

// Phase 3 of PBA. m3 must divides texture size
// This method color along the Y axis
void pba3DColorYAxis(int m3) 
{
	dim3 block = dim3(BLOCKX, m3); 
    dim3 grid = dim3(pbaTexSize / block.x, pbaTexSize); 

    cudaBindTexture(0, pbaTexColor, pbaTextures[1 - pbaCurrentBuffer]); 
    kernelColorAxis<<< grid, block >>>(pbaTextures[pbaCurrentBuffer], pbaTexSize); 
    cudaUnbindTexture(pbaTexColor); 
}

void pba3DCompute(int m1, int m2, int m3)
{
    /************* Compute Z axis *************/
    // --> (X, Y, Z)
	pba3DColorZAxis(m1); 

	/************* Compute Y axis *************/
    // --> (X, Y, Z)
	pba3DComputeProximatePointsYAxis(m2);
	pba3DColorYAxis(m3); 

    // --> (Y, X, Z)
    pba3DTransposeXY(pbaTextures[pbaCurrentBuffer]); 

    /************** Compute X axis *************/
    // Compute X axis
	pba3DComputeProximatePointsYAxis(m2);
	pba3DColorYAxis(m3); 

    // --> (X, Y, Z)
    pba3DTransposeXY(pbaTextures[pbaCurrentBuffer]); 
}

// Compute 3D Voronoi diagram
// Input: a 3D texture. Each pixel is an integer encoding 3 coordinates. 
//    For each site at (x, y, z), the pixel at coordinate (x, y, z) should contain 
//    the encoded coordinate (x, y, z). Pixels that are not sites should contain 
//    the integer MARKER. Use ENCODE (and DECODE) macro to encode (and decode).
// See original paper for the effect of the three parameters: 
//     phase1Band, phase2Band, phase3Band
// Parameters must divide textureSize
void pba3DVoronoiDiagram(int *input, int *output, 
                         int phase1Band, int phase2Band, int phase3Band) 
{
	// Initialization
	pba3DInitializeInput(input); 
	
    // Compute the 3D Voronoi Diagram
    pba3DCompute(phase1Band, phase2Band, phase3Band); 

    // Copy back the result
    cudaMemcpy(output, pbaTextures[pbaCurrentBuffer], pbaMemSize, cudaMemcpyDeviceToHost); 
}

